unit hiMutexThread;

interface

{$I share.inc}

uses Kol,Share,Windows,Debug;
var
  MutexHandle: THandle;
  
type
  THIMutexThread = class(TDebug)
   private
    th:PThread;
    FStop:boolean;

    function Execute(Sender:PThread): Integer;
    procedure SyncExec;
    function waitqueue(Delay:cardinal; Single:boolean=false):DWord;
    procedure endprotect;

   public
    _prop_Delay:cardinal;
    _prop_FastStop:boolean;
    _prop_BusyEvent:byte;

    _event_onSyncExec:THI_Event;
    _event_onExec:THI_Event;
    _event_onError:THI_Event;
    _event_onWaitqueue:THI_Event;    

    destructor Destroy; override;
    procedure _work_doStart(var _Data:TData; Index:word);
    procedure _work_doStop(var _Data:TData; Index:word);
    procedure _work_doStopFlag(var _Data:TData; Index:word);
    procedure _work_doDelay(var _Data:TData; Index:word);
    procedure _var_Busy(var _Data:TData; Index:word);
  end;

implementation

destructor THIMutexThread.Destroy;
var dt:TData;
begin
  _work_doStop(dt,0);
  inherited;
end;

function THIMutexThread.Execute(Sender:PThread): Integer;
begin
  FStop := false;
  repeat
    if waitqueue(_prop_delay, true) <> WAIT_OBJECT_0 then
    begin
      _hi_OnEvent(_event_onWaitqueue);
      waitqueue(_prop_delay);
    end;
    _hi_OnEvent(_event_onExec);
    if Assigned(_event_onSyncExec.Event) then Sender.Synchronize(SyncExec);
    endprotect;
  until Sender.Terminated or _prop_FastStop or FStop;
  Sender.Tag := 0;
  Result := 0;
end;

procedure THIMutexThread.SyncExec;
begin
  _hi_OnEvent(_event_onSyncExec);
end;

procedure THIMutexThread._work_doStart;
begin
  if Assigned(th) then
  begin
    if (_prop_BusyEvent = 0) and (th.Tag = 1) then exit;
    th.Terminate;
    th.WaitFor;
    free_and_nil(th);
  end;
  th := {$ifdef F_P}NewThreadforFPC{$else}NewThread{$endif};
  th.OnExecute := Execute;
  th.Resume;
  th.Tag := 1;
end;

procedure THIMutexThread._work_doStop;
begin
  if Assigned(th) then
  begin
    th.Terminate;
    th.WaitFor;
    free_and_nil(th);
  end;
end;

procedure THIMutexThread._work_doStopFlag;
begin
  FStop := true;
end;

procedure THIMutexThread._work_doDelay;
begin
  _prop_Delay := ToInteger(_Data);
end;

procedure THIMutexThread._var_Busy;
begin
  dtInteger(_Data, integer(Assigned(th) and (th.Tag = 1)));
end;

function THIMutexThread.waitqueue;
begin
  repeat
    Sleep(Delay);
    Result := WaitForSingleObject(MutexHandle, 0);
    if Single then break;
  until Result = WAIT_OBJECT_0;  
end;

procedure THIMutexThread.endprotect;
begin
  ReleaseMutex(MutexHandle);
end;

initialization
  MutexHandle := CreateMutex(nil, false, PChar(int2str(GetCurrentProcessID)));

finalization
  if MutexHandle <> 0 then CloseHandle(MutexHandle);

end.